@@ -13,23 +13,26 @@
 # limitations under the License.
 # ==============================================================================
 
-from distutils.version import LooseVersion
+import os
+from packaging import version
 
 import horovod.tensorflow as hvd
 import tensorflow as tf
 from horovod.tensorflow.gradient_aggregation import LocalGradientAggregationHelper
 from horovod.tensorflow.gradient_aggregation_eager import LocalGradientAggregationHelperEager
-from horovod.tensorflow.mpi_ops import rank
+from horovod.tensorflow.mpi_ops import rank, size_op
 
 
-_PRE_TF_2_4_0 = LooseVersion(tf.__version__) < LooseVersion('2.4.0')
+_PRE_TF_2_4_0 = version.parse(tf.__version__) < version.parse('2.4.0')
+_IS_TF2 = version.parse(tf.__version__) >= version.parse('2.0.0')
 
 
 def create_distributed_optimizer(keras, optimizer, name, device_dense, device_sparse,
                                  compression, sparse_as_dense, gradient_predivide_factor,
                                  op, backward_passes_per_step=1,
                                  average_aggregated_gradients=False,
-                                 groups=None, process_set=hvd.global_process_set):
+                                 groups=None, process_set=hvd.global_process_set,
+                                 scale_local_gradients=True):
     class _DistributedOptimizer(keras.optimizers.Optimizer):
         _HAS_AGGREGATE_GRAD = True
 
@@ -50,6 +53,9 @@ def create_distributed_optimizer(keras, optimizer, name, device_dense, device_sp
                 groups,
                 process_set=process_set)
 
+            self._local_vars = set()
+            self.process_set = process_set
+            self.scale_local_gradients = scale_local_gradients
             self._agg_helper = None
             if backward_passes_per_step > 1:
                 if hvd._executing_eagerly():
@@ -58,6 +64,8 @@ def create_distributed_optimizer(keras, optimizer, name, device_dense, device_sp
                         allreduce_func=self._allreduce_grads,
                         sparse_as_dense=sparse_as_dense,
                         average_aggregated_gradients=average_aggregated_gradients,
+                        process_set=process_set,
+                        scale_local_gradients=scale_local_gradients
                     )
                 else:
                     self._agg_helper = LocalGradientAggregationHelper(
@@ -67,8 +75,21 @@ def create_distributed_optimizer(keras, optimizer, name, device_dense, device_sp
                         average_aggregated_gradients=average_aggregated_gradients,
                         rank=rank(),
                         optimizer_type=LocalGradientAggregationHelper._OPTIMIZER_TYPE_KERAS,
+                        process_set=process_set,
+                        scale_local_gradients=scale_local_gradients
                     )
 
+        def register_local_var(self, var):
+            """Registers a source/variable as worker local. Horovod will not perform any global
+            operations on gradients corresponding to these sources and will instead return the local
+            gradient."""
+            if self._agg_helper:
+                self._agg_helper.register_local_var(var)
+            elif _IS_TF2:
+                self._local_vars.add(var.ref())
+            else:
+                self._local_vars.add(var)
+
         def _compute_gradients(self, loss, var_list, grad_loss=None, tape=None):
             """
             Compute gradients of all trainable variables.
@@ -121,7 +142,47 @@ def create_distributed_optimizer(keras, optimizer, name, device_dense, device_sp
             if self._agg_helper:
                 return self._agg_helper.compute_gradients(tuple(grads), tuple(vars))
             else:
-                return self._allreduce_grads(grads, vars)
+                def __filtered_reduce_grads(grads, vars):
+                    rv = []
+                    rg = []
+                    if _IS_TF2:
+                        v2g = {var.ref(): grad for var, grad in zip(vars, grads)}
+                        for var, grad in zip(vars, grads):
+                            if var.ref() not in self._local_vars:
+                                rv.append(var)
+                                rg.append(grad)
+                    else:
+                        v2g = {var: grad for var, grad in zip(vars, grads)}
+                        for var, grad in zip(vars, grads):
+                            if var not in self._local_vars:
+                                rv.append(var)
+                                rg.append(grad)
+
+                    rg = self._allreduce_grads(rg, rv)
+                    horovod_size = size_op(process_set_id=self.process_set.process_set_id) if int(os.environ.get("HOROVOD_ELASTIC", 0)) else self.process_set.size()
+                    if _IS_TF2:
+                        for rv, rg in zip(rv, rg):
+                            v2g[rv.ref()] = rg
+
+                        if self.scale_local_gradients and len(self._local_vars):
+                            # Scale local gradients by a size factor. See pull/3695 and discussions/3705 for context.
+                            for v_ref in v2g:
+                                if v_ref in self._local_vars and v2g[v_ref] is not None:
+                                    v2g[v_ref] /= horovod_size
+
+                        return [v2g[rv.ref()] for rv in vars]
+                    else:
+                        for rv, rg in zip(rv, rg):
+                            v2g[rv] = rg
+
+                        if self.scale_local_gradients and len(self._local_vars):
+                            # Scale local gradients by a size factor. See pull/3695 and discussions/3705 for context.
+                            for v in v2g:
+                                if v in self._local_vars and v2g[v] is not None:
+                                    v2g[v] /= horovod_size
+
+                        return [v2g[rv] for rv in vars]
+                return __filtered_reduce_grads(grads, vars)
 
         def apply_gradients(self, *args, **kwargs):
             if self._agg_helper:
