@@ -29,6 +29,8 @@
 
 #if HAVE_CUDA
 #include "ops/cuda/cuda_kernels.h"
+#elif HAVE_ROCM
+#include "ops/rocm/hip_kernels.h"
 #endif
 
 
@@ -598,6 +600,28 @@ Response Controller::ConstructResponse(const std::string& name, int joined_size)
     }
   }
 
+  // If we are doing an allreduce, check that the reduction op is indentical
+  // across ranks.
+  ReduceOp reduce_op;
+  if (message_type == Request::ALLREDUCE) {
+    reduce_op = requests[0].reduce_op();
+    for (unsigned int i = 1; i < requests.size(); ++i) {
+      if (error) {
+        break;
+      }
+      ReduceOp request_reduce_op = requests[i].reduce_op();
+      if (reduce_op != request_reduce_op) {
+        error = true;
+        error_message_stream
+            << "Mismatched reduction operation: "
+            << "One rank sent reduction op " << reduce_op
+            << ", but another rank sent reduction op "
+            << request_reduce_op << ".";
+        break;
+      }
+    }
+  }
+
   std::vector<int64_t> tensor_sizes;
   if (message_type == Request::ALLGATHER ||
       message_type == Request::ALLTOALL) {
@@ -684,7 +708,14 @@ Response Controller::ConstructResponse(const std::string& name, int joined_size)
     for (auto dim : requests[0].tensor_shape()) {
       tensor_shape.AddDim(dim);
     }
-    tensor_sizes.push_back(tensor_shape.num_elements());
+    if (tensor_shape.dims() == 0) {
+      error = true;
+      error_message_stream << "Process-set rank zero tried to "
+                           << Request::RequestType_Name(message_type)
+                           << " a rank-zero tensor.";
+    } else {
+      tensor_sizes.push_back(tensor_shape.num_elements());
+    }
   }
 
   if (message_type == Request::ALLREDUCE || message_type == Request::ADASUM) {
@@ -766,6 +797,7 @@ Response Controller::ConstructResponse(const std::string& name, int joined_size)
     response.set_tensor_type(data_type);
     response.set_prescale_factor(prescale_factor);
     response.set_postscale_factor(postscale_factor);
+    response.set_reduce_op(reduce_op);
   } else if (message_type == Request::BROADCAST) {
     response.set_response_type(Response::BROADCAST);
   } else if (message_type == Request::ALLTOALL) {
@@ -840,12 +872,15 @@ void Controller::FuseResponses(std::deque<Response>& responses,
       // Attempt to add more responses to this fused response.
 
       tensor_size = response.tensor_sizes()[0] * GetTypeSize(response.tensor_type());
-#if HAVE_CUDA
-      if (state.batch_d2d_memcopies) {
+#if HAVE_CUDA || HAVE_ROCM
+      if (state.batch_d2d_memcopies &&
+          response.response_type() != Response::ResponseType::REDUCESCATTER) {
         // Add 16 byte pad for batched memcpy op
-        tensor_size = BATCHED_D2D_PADDING * ((tensor_size + BATCHED_D2D_PADDING - 1) / BATCHED_D2D_PADDING);
+        tensor_size =
+            BATCHED_D2D_PADDING *
+            ((tensor_size + BATCHED_D2D_PADDING - 1) / BATCHED_D2D_PADDING);
       }
-#endif
+#endif // HAVE_CUDA || HAVE_ROCM
       std::deque<Response> skipped_responses;
       int64_t skipped_size = 0;
       while (!responses.empty()) {
@@ -857,19 +892,23 @@ void Controller::FuseResponses(std::deque<Response>& responses,
                                       : new_response.tensor_sizes()[0] *
                                         GetTypeSize(new_response.tensor_type());
 
-#if HAVE_CUDA
-        if (state.batch_d2d_memcopies) {
+#if HAVE_CUDA || HAVE_ROCM
+        if (state.batch_d2d_memcopies &&
+            response.response_type() != Response::ResponseType::REDUCESCATTER) {
           // Add 16 byte pad for batched memcpy op
-          new_tensor_size = BATCHED_D2D_PADDING * ((new_tensor_size + BATCHED_D2D_PADDING - 1) / BATCHED_D2D_PADDING);
+          new_tensor_size = BATCHED_D2D_PADDING *
+                            ((new_tensor_size + BATCHED_D2D_PADDING - 1) /
+                             BATCHED_D2D_PADDING);
         }
-#endif
+#endif // HAVE_CUDA || HAVE_ROCM
 
         if (response.response_type() == new_response.response_type() &&
             response.devices() == new_response.devices() &&
             response.tensor_type() == new_response.tensor_type() &&
             tensor_size + new_tensor_size <= TensorFusionThresholdBytes() &&
             response.prescale_factor() == new_response.prescale_factor() &&
-            response.postscale_factor() == new_response.postscale_factor()) {
+            response.postscale_factor() == new_response.postscale_factor() &&
+            response.reduce_op() == new_response.reduce_op()) {
           // These tensors will fuse together well.
           tensor_size += new_tensor_size;
           response.add_tensor_name(std::move(new_response.tensor_names()[0]));
