@@ -13,6 +13,9 @@
 // limitations under the License.
 // =============================================================================
 
+// ATTENTION: Any change here might obsolete hip_kernels.h in rocm folder.
+//            Please keep this file synced with hip_kernels.h.
+
 #ifndef CUDA_KERNELS_H
 #define CUDA_KERNELS_H
 
@@ -39,6 +42,9 @@ void BatchedD2DMemcpyCudaImpl(BatchedD2DParams& params, int num_copies, cudaStre
 void ScaleBufferCudaImpl(const void* fused_input_data, void* buffer_data, const int64_t num_elements,
                          double scale_factor, DataType dtype, cudaStream_t stream);
 
+void BatchedScaledD2DMemcpyCudaImpl(BatchedD2DParams& params, int num_copies, double scale_factor,
+                                    DataType dtype, cudaStream_t stream);
+
 } // namespace common
 } // namespace horovod
 
