@@ -188,7 +188,7 @@ protected:
   void ScaleBuffer(double scale_factor,
                    const std::vector<TensorTableEntry>& entries,
                    const void* fused_input_data, void* buffer_data,
-                   int64_t num_elements);
+                   int64_t num_elements) override;
 
   GPUContext* gpu_context_;
   GPUOpContext gpu_op_context_;
@@ -203,6 +203,16 @@ public:
                const Response& response) const override;
 
 protected:
+#if HAVE_GPU
+  void MemcpyInFusionBuffer(const std::vector<TensorTableEntry>& entries,
+                            const int* displcmnts, int element_size,
+                            void*& buffer_data) override;
+
+  void MemcpyOutFusionBuffer(const int64_t* const* entry_component_offsets,
+                             const int64_t* const* entry_component_sizes,
+                             const void* buffer_data, int element_size,
+                             std::vector<TensorTableEntry>& entries) override;
+#endif
   void MemcpyEntryInFusionBuffer(const std::vector<TensorTableEntry>& entries,
                                  const TensorTableEntry& e,
                                  void* buffer_data_at_offset) override;
@@ -225,7 +235,7 @@ public:
                const Response& response) const override;
 
 protected:
-  struct GPUContext* gpu_context_;
+  GPUContext* gpu_context_;
   GPUOpContext gpu_op_context_;
 };
 
@@ -241,9 +251,9 @@ protected:
   GPUOpContext gpu_op_context_;
 };
 
-class GPUReduceScatter : public ReducescatterOp {
+class GPUReducescatter : public ReducescatterOp {
 public:
-  GPUReduceScatter(GPUContext* context, HorovodGlobalState* global_state);
+  GPUReducescatter(GPUContext* context, HorovodGlobalState* global_state);
 
   bool Enabled(const ParameterManager& param_manager,
                const std::vector<TensorTableEntry>& entries,
@@ -257,6 +267,16 @@ protected:
   void MemcpyEntryOutFusionBuffer(const void* buffer_data_at_offset,
                                   TensorTableEntry& e) override;
 
+#if HAVE_GPU
+  void MemcpyInFusionBuffer(
+      const std::vector<TensorTableEntry>& entries,
+      const std::vector<std::vector<TensorShape>>& output_shapes,
+      std::size_t element_size, void*& buffer_data) override;
+
+  void MemcpyOutFusionBuffer(const void* buffer_data,
+                             std::vector<TensorTableEntry>& entries) override;
+#endif // HAVE_GPU
+
   GPUContext* gpu_context_;
   GPUOpContext gpu_op_context_;
 };
