@@ -23,8 +23,12 @@
 #if NCCL_VERSION_CODE >= NCCL_VERSION(2, 7, 0)
 #define NCCL_P2P_SUPPORTED
 #endif
+#if NCCL_VERSION_CODE >= NCCL_VERSION(2, 10, 0)
+#define NCCL_AVG_SUPPORTED
+#endif
 #elif HAVE_ROCM
 #include <rccl.h>
+#define NCCL_P2P_SUPPORTED
 #endif
 
 #include "../hashes.h"
@@ -180,6 +184,9 @@ protected:
     std::shared_ptr<ReadyEvent> event;
     Status status = e.context->AllocateOutput(output_shape, &e.output, &event);
     if (!status.ok()) {
+      LOG(WARNING)
+          << "NCCLAlltoall::PrepareOutputAndParams failed to allocate output: "
+          << status.reason();
       return status;
     }
 
@@ -197,6 +204,9 @@ protected:
                                                &e.received_splits,
                                                &revent);
     if (!rstatus.ok()) {
+      LOG(WARNING) << "NCCLAlltoall::PrepareOutputAndParams failed to allocate "
+                      "received_splits: "
+                   << status.reason();
       return rstatus;
     }
 
@@ -240,6 +250,34 @@ private:
 };
 #endif
 
+class NCCLTorusAllreduce : public GPUAllreduce {
+public:
+  NCCLTorusAllreduce(NCCLContext* local_nccl_context, NCCLContext* cross_nccl_context,
+                     GPUContext* gpu_context, HorovodGlobalState* global_state)
+      : GPUAllreduce(gpu_context, global_state),
+        local_nccl_context_(local_nccl_context),
+        cross_nccl_context_(cross_nccl_context),
+        local_nccl_op_context_(local_nccl_context, global_state, Communicator::LOCAL),
+        cross_nccl_op_context_(cross_nccl_context, global_state, Communicator::CROSS),
+        global_state_(global_state){};
+
+  Status Execute(std::vector<TensorTableEntry>& entries,
+                 const Response& response) override;
+
+  bool Enabled(const ParameterManager& param_manager,
+               const std::vector<TensorTableEntry>& entries,
+               const Response& response) const override;
+
+protected:
+  void WaitForData(std::vector<TensorTableEntry>& entries) override;
+
+  NCCLContext* local_nccl_context_;
+  NCCLOpContext local_nccl_op_context_;
+  NCCLContext* cross_nccl_context_;
+  NCCLOpContext cross_nccl_op_context_;
+  HorovodGlobalState* global_state_;
+};
+
 class NCCLAllgather : public GPUAllgather {
 public:
   NCCLAllgather(NCCLContext* nccl_context, GPUContext* gpu_context,
@@ -268,11 +306,11 @@ protected:
   HorovodGlobalState* global_state_;
 };
 
-class NCCLReducescatter : public GPUReduceScatter {
+class NCCLReducescatter : public GPUReducescatter {
 public:
   NCCLReducescatter(NCCLContext* nccl_context, GPUContext* gpu_context,
                     HorovodGlobalState* global_state)
-      : GPUReduceScatter(gpu_context, global_state),
+      : GPUReducescatter(gpu_context, global_state),
         nccl_context_(nccl_context),
         nccl_op_context_(nccl_context, global_state, Communicator::GLOBAL),
         global_state_(global_state) {}
@@ -286,7 +324,7 @@ public:
 
 protected:
   Status AllocateOutput(std::vector<TensorTableEntry>& entries,
-                                const std::vector<TensorShape>& output_shapes) override;
+                        const std::vector<TensorShape>& output_shapes) override;
 
   void WaitForData(std::vector<TensorTableEntry>& entries) override;
 
