@@ -38,7 +38,7 @@ Status TensorQueue::AddToTensorQueueMulti(std::vector<TensorTableEntry>& entries
                                           std::vector<Request>& messages) {
   std::lock_guard<std::mutex> guard(mutex_);
 
-  for (int i = 0; i < entries.size(); ++i) {
+  for (std::size_t i = 0; i < entries.size(); ++i) {
     if (tensor_table_.find(entries[i].tensor_name) != tensor_table_.end()) {
       return DUPLICATE_NAME_ERROR;
     }
@@ -90,24 +90,38 @@ void TensorQueue::GetTensorEntriesFromResponse(
   {
     // Lock on the tensor table.
     std::lock_guard<std::mutex> guard(mutex_);
+    if (response.response_type() == Response::JOIN) {
+      assert(response.tensor_names().size() == 1);
+      assert(response.tensor_names()[0] == JOIN_TENSOR_NAME);
+      auto iter = tensor_table_.find(JOIN_TENSOR_NAME);
+      assert(iter != tensor_table_.end());
+
+      entries.push_back(std::move(iter->second));
+
+      // The tensor table will be cleared of the join tensor later in
+      // RemoveJoinTensor().
+      return;
+    }
     int64_t i = 0;
     for (auto& name : response.tensor_names()) {
       assert(response.response_type() == Response::ALLREDUCE ||
              response.response_type() == Response::ALLGATHER ||
              response.response_type() == Response::BROADCAST ||
              response.response_type() == Response::ALLTOALL ||
+             response.response_type() == Response::REDUCESCATTER ||
              response.response_type() == Response::ADASUM ||
+             response.response_type() == Response::BARRIER ||
              response.response_type() == Response::ERROR);
 
       if (!joined) {
         // We should never fail at finding this key in the tensor table.
         auto iter = tensor_table_.find(name);
         assert(iter != tensor_table_.end());
-
         entries.push_back(std::move(iter->second));
 
         // Clear the tensor table of this tensor.
         tensor_table_.erase(iter);
+
       } else if (response.response_type() != Response::ERROR) {
 
         // Find Join tensor to use its context.
@@ -140,6 +154,12 @@ TensorQueue::GetTensorEntry(const std::string& tensor_name) const{
   return iter;
 }
 
+bool TensorQueue::IsTensorPresentInTable(const std::string& tensor_name) const {
+  // Lock on the tensor table.
+  std::lock_guard<std::mutex> guard(mutex_);
+  return tensor_table_.find(tensor_name) != tensor_table_.end();
+}
+
 // Pop out all the messages from the queue
 void TensorQueue::PopMessagesFromQueue(
     std::deque<Request>& message_queue_buffer) {
