@@ -145,7 +145,7 @@ void DoHorovodOperation(void*, void* on_complete_ptr, void* param) {
   auto ops_param = static_cast<MpiOpsParam*>(param);
   auto first_tensor = ops_param->input_tensors[0].get();
   auto device = TensorUtil::GetDevice(first_tensor);
-  auto average = ops_param->average;
+  auto reduce_op = ops_param->reduce_op;
   auto prescale_factor = ops_param->prescale_factor;
   auto postscale_factor = ops_param->postscale_factor;
   auto num_tensors = ops_param->input_tensors.size();
@@ -172,7 +172,16 @@ void DoHorovodOperation(void*, void* on_complete_ptr, void* param) {
     if (TensorUtil::GetDevice(input_tensor) != device) {
       throw std::logic_error("Tensors in list must be on same device.");
     }
-    auto ctx = std::make_shared<MXOpContext>(device, output);
+    std::shared_ptr<MXOpContext> ctx;
+    if (ops_param->op_type == OperationType::ALLGATHER ||
+        ops_param->op_type == OperationType::REDUCESCATTER) {
+      // Grouped allgather and reducescatter operations will need to allocate
+      // the correct output tensor later on, therefore referring to all output
+      // tensors in context.
+      ctx = std::make_shared<MXOpContext>(device, ops_param->outputs);
+    } else {
+      ctx = std::make_shared<MXOpContext>(device, output);
+    }
     if (ops_param->received_splits_tensor) {
       ctx->AddOutput(ops_param->received_splits_tensor.get());
     }
@@ -256,13 +265,13 @@ void DoHorovodOperation(void*, void* on_complete_ptr, void* param) {
       enqueue_result = EnqueueTensorAllreduces(
           hvd_contexts, hvd_tensors, hvd_outputs, ready_event_lists,
           ops_param->op_names, device, callbacks,
-          (average) ? ReduceOp::AVERAGE : ReduceOp::SUM, prescale_factor,
+          reduce_op, prescale_factor,
           postscale_factor, process_set_id);
       break;
     case OperationType::ALLGATHER:
-      enqueue_result = EnqueueTensorAllgather(
-          hvd_contexts[0], hvd_tensors[0], ready_event_lists[0],
-          ops_param->op_names[0], device, callbacks[0], process_set_id);
+      enqueue_result = EnqueueTensorAllgathers(
+          hvd_contexts, hvd_tensors, ready_event_lists,
+          ops_param->op_names, device, callbacks, process_set_id);
       break;
     case OperationType::BROADCAST:
       if (horovod_rank() != ops_param->root_rank) {
@@ -288,9 +297,9 @@ void DoHorovodOperation(void*, void* on_complete_ptr, void* param) {
       break;
     }
     case OperationType::REDUCESCATTER:
-      enqueue_result = EnqueueTensorReducescatter(
-          hvd_contexts[0], hvd_tensors[0], ready_event_lists[0],
-          ops_param->op_names[0], device, callbacks[0],
+      enqueue_result = EnqueueTensorReducescatters(
+          hvd_contexts, hvd_tensors, ready_event_lists,
+          ops_param->op_names, device, callbacks,
           ReduceOp::SUM, process_set_id);
       break;
     default:
@@ -305,7 +314,7 @@ inline void PushHorovodOperation(OperationType op_type, NDArray* const * inputs,
                                  int priority, int num_tensors,
                                  int process_set_id = 0,
                                  int root_rank = -1,
-                                 bool average = true,
+                                 ReduceOp reduce_op = ReduceOp::AVERAGE,
                                  NDArray* splits = nullptr,
                                  NDArray* output_received_splits = nullptr,
                                  double prescale_factor = 1.0,
@@ -380,7 +389,7 @@ inline void PushHorovodOperation(OperationType op_type, NDArray* const * inputs,
   auto ops_param = CreateMpiOpsParam(
       std::move(input_copies), std::move(output_copies), std::move(outputs_vec),
       cpu_input_tensors, cpu_output_tensors, op_type, std::move(op_names),
-      root_rank, average, splits_tensor, received_splits_tensor,
+      root_rank, reduce_op, splits_tensor, received_splits_tensor,
       prescale_factor, postscale_factor, process_set_id);
 
   // Not in-place
@@ -416,7 +425,7 @@ void DoHorovodOperationCudaOnCPU(void*, void* on_complete_ptr, void* param) {
   auto on_complete = *static_cast<CallbackOnComplete*>(on_complete_ptr);
   auto ops_param = static_cast<MpiOpsParam*>(param);
   auto device = CPU_DEVICE_ID;
-  auto average = ops_param->average;
+  auto reduce_op = ops_param->reduce_op;
   auto prescale_factor = ops_param->prescale_factor;
   auto postscale_factor = ops_param->postscale_factor;
   auto num_tensors = ops_param->cpu_input_tensors.size();
@@ -426,15 +435,20 @@ void DoHorovodOperationCudaOnCPU(void*, void* on_complete_ptr, void* param) {
   std::vector<std::shared_ptr<OpContext>> hvd_contexts;
   std::vector<StatusCallback> callbacks;
   std::vector<ReadyEventList> ready_event_lists;
+  std::vector<NDArray *> outputs;
   hvd_cpu_buffers.reserve(num_tensors);
   hvd_contexts.reserve(num_tensors);
   ready_event_lists.reserve(num_tensors);
   callbacks.reserve(num_tensors);
+  outputs.reserve(num_tensors);
+  for (std::size_t i = 0; i < num_tensors; ++i) {
+    outputs.push_back(ops_param->cpu_output_tensors[i].get());
+  }
 
   auto callback_mutex = std::make_shared<std::mutex>();
   for (std::size_t i = 0; i < num_tensors; ++i) {
     auto input = ops_param->cpu_input_tensors[i].get();
-    auto output = ops_param->cpu_output_tensors[i].get();
+    auto output = outputs[i];
 
     ready_event_lists.emplace_back(FormReadyEventList(input, output));
 
@@ -442,7 +456,16 @@ void DoHorovodOperationCudaOnCPU(void*, void* on_complete_ptr, void* param) {
     if (TensorUtil::GetDevice(input) != device) {
       throw std::logic_error("Tensors in list must be on same device.");
     }
-    auto ctx = std::make_shared<MXOpContext>(device, output);
+    std::shared_ptr<MXOpContext> ctx;
+    if (ops_param->op_type == OperationType::ALLGATHER ||
+        ops_param->op_type == OperationType::REDUCESCATTER) {
+      // Grouped allgather and reducescatter operations will need to allocate
+      // the correct output tensor later on, therefore referring to all output
+      // tensors in context.
+      ctx = std::make_shared<MXOpContext>(device, outputs);
+    } else {
+      ctx = std::make_shared<MXOpContext>(device, output);
+    }
     if (ops_param->received_splits_tensor) {
       ctx->AddOutput(ops_param->received_splits_tensor.get());
     }
@@ -467,13 +490,13 @@ void DoHorovodOperationCudaOnCPU(void*, void* on_complete_ptr, void* param) {
     enqueue_result = EnqueueTensorAllreduces(
         hvd_contexts, hvd_cpu_buffers, hvd_cpu_buffers, ready_event_lists,
         ops_param->op_names, device, callbacks,
-        (average) ? ReduceOp::AVERAGE : ReduceOp::SUM, prescale_factor,
+        reduce_op, prescale_factor,
         postscale_factor, process_set_id);
     break;
   case OperationType::ALLGATHER:
-    enqueue_result = EnqueueTensorAllgather(
-        hvd_contexts[0], hvd_cpu_buffers[0], ready_event_lists[0],
-        ops_param->op_names[0], device, callbacks[0], process_set_id);
+    enqueue_result = EnqueueTensorAllgathers(
+        hvd_contexts, hvd_cpu_buffers, ready_event_lists,
+        ops_param->op_names, device, callbacks, process_set_id);
     break;
   case OperationType::BROADCAST:
     enqueue_result = EnqueueTensorBroadcast(
@@ -490,9 +513,9 @@ void DoHorovodOperationCudaOnCPU(void*, void* on_complete_ptr, void* param) {
     break;
   }
   case OperationType::REDUCESCATTER:
-    enqueue_result = EnqueueTensorReducescatter(
-        hvd_contexts[0], hvd_cpu_buffers[0], ready_event_lists[0],
-        ops_param->op_names[0], device, callbacks[0],
+    enqueue_result = EnqueueTensorReducescatters(
+        hvd_contexts, hvd_cpu_buffers, ready_event_lists,
+        ops_param->op_names, device, callbacks,
         ReduceOp::SUM, process_set_id);
     break;
   default:
@@ -507,7 +530,7 @@ inline void PushHorovodOperationCudaOnCPU(OperationType op_type, NDArray* const
                                           int priority, int num_tensors,
                                           int process_set_id = 0,
                                           int root_rank = -1,
-                                          bool average = true,
+                                          ReduceOp reduce_op = ReduceOp::AVERAGE,
                                           NDArray* splits = nullptr,
                                           NDArray* output_received_splits = nullptr,
                                           double prescale_factor = 1.0,
@@ -564,7 +587,7 @@ inline void PushHorovodOperationCudaOnCPU(OperationType op_type, NDArray* const
   auto ops_param = CreateMpiOpsParam(
       std::move(input_copies), std::move(output_copies), std::move(outputs_vec),
       cpu_input_tensors, cpu_output_tensors, op_type, std::move(op_names),
-      root_rank, average, splits_tensor, received_splits_tensor,
+      root_rank, reduce_op, splits_tensor, received_splits_tensor,
       prescale_factor, postscale_factor, process_set_id);
 
   std::vector<void*> cpu_input_vars;
@@ -614,7 +637,7 @@ inline void PushHorovodOperationCudaOnCPU(OperationType op_type, NDArray* const
 
 extern "C" int horovod_mxnet_allreduce_async(NDArray* const * inputs,
                                              NDArray* const * outputs,
-                                             const char* name, bool average,
+                                             const char* name, int reduce_op_int,
                                              int priority,
                                              double prescale_factor,
                                              double postscale_factor,
@@ -622,27 +645,29 @@ extern "C" int horovod_mxnet_allreduce_async(NDArray* const * inputs,
                                              int process_set_id) {
   MX_API_BEGIN();
 
+  ReduceOp reduce_op = static_cast<ReduceOp>(reduce_op_int);
+
 #if HAVE_ROCM
   // Averaging left at framework level for ROCm until ScaleBuffer implementation
   // added.
-  bool average_in_framework = average;
-  average = false;
+  bool average_in_framework = (reduce_op == ReduceOp::AVERAGE);
+  reduce_op = ReduceOp::SUM;
 #endif
 
 #if HAVE_CUDA && !HOROVOD_GPU_ALLREDUCE
   if (IsTensorOnCPU(inputs[0]) && IsTensorOnCPU(outputs[0])) {
     PushHorovodOperation(OperationType::ALLREDUCE, inputs, outputs, name,
-                         priority, num_tensors, process_set_id, -1, average,
+                         priority, num_tensors, process_set_id, -1, reduce_op,
                          nullptr, nullptr, prescale_factor, postscale_factor);
   } else {
     PushHorovodOperationCudaOnCPU(OperationType::ALLREDUCE, inputs, outputs,
                                   name, priority, num_tensors, process_set_id,
-                                  -1, average, nullptr, nullptr,
+                                  -1, reduce_op, nullptr, nullptr,
                                   prescale_factor, postscale_factor);
   }
 #else
   PushHorovodOperation(OperationType::ALLREDUCE, inputs, outputs, name,
-                       priority, num_tensors, process_set_id, -1, average,
+                       priority, num_tensors, process_set_id, -1, reduce_op,
                        nullptr, nullptr, prescale_factor, postscale_factor);
 #endif
 
@@ -657,23 +682,24 @@ extern "C" int horovod_mxnet_allreduce_async(NDArray* const * inputs,
   MX_API_END();
 }
 
-extern "C" int horovod_mxnet_allgather_async(NDArray* input,
-                                             NDArray* output,
+extern "C" int horovod_mxnet_allgather_async(NDArray* const* inputs,
+                                             NDArray* const* outputs,
                                              const char* name, int priority,
-                                             int process_set_id) {
+                                             int process_set_id,
+                                             int num_tensors) {
   MX_API_BEGIN();
 
 #if HAVE_CUDA && !HOROVOD_GPU_ALLGATHER
-  if (IsTensorOnCPU(input) && IsTensorOnCPU(output)) {
-    PushHorovodOperation(OperationType::ALLGATHER, &input, &output,
-                         name, priority, 1, process_set_id);
+  if (IsTensorOnCPU(inputs[0]) && IsTensorOnCPU(outputs[0])) {
+    PushHorovodOperation(OperationType::ALLGATHER, inputs, outputs,
+                         name, priority, num_tensors, process_set_id);
   } else {
-    PushHorovodOperationCudaOnCPU(OperationType::ALLGATHER, &input, &output,
-                                  name, priority, 1, process_set_id);
+    PushHorovodOperationCudaOnCPU(OperationType::ALLGATHER, inputs, outputs,
+                                  name, priority, num_tensors, process_set_id);
   }
 #else
-  PushHorovodOperation(OperationType::ALLGATHER, &input, &output,
-                       name, priority, 1, process_set_id);
+  PushHorovodOperation(OperationType::ALLGATHER, inputs, outputs,
+                       name, priority, num_tensors, process_set_id);
 #endif
 
   MX_API_END();
@@ -715,40 +741,41 @@ extern "C" int horovod_mxnet_alltoall_async(NDArray* input,
 #if HAVE_CUDA && !HOROVOD_GPU_ALLTOALL
   if (IsTensorOnCPU(input) && IsTensorOnCPU(output)) {
     PushHorovodOperation(OperationType::ALLTOALL, &input, &output, name,
-                         priority, 1, process_set_id, -1, false, splits,
+                         priority, 1, process_set_id, -1, ReduceOp::SUM, splits,
                          output_received_splits);
 
   } else {
     PushHorovodOperationCudaOnCPU(OperationType::ALLTOALL, &input, &output,
-                                  name, priority, 1, process_set_id, -1, false,
+                                  name, priority, 1, process_set_id, -1, ReduceOp::SUM,
                                   splits, output_received_splits);
   }
 #else
   PushHorovodOperation(OperationType::ALLTOALL, &input, &output, name, priority,
-                       1, process_set_id, -1, false, splits,
+                       1, process_set_id, -1, ReduceOp::SUM, splits,
                        output_received_splits);
 #endif
 
   MX_API_END();
 }
 
-extern "C" int
-horovod_mxnet_reducescatter_async(NDArray* input, NDArray* output,
-                                  const char* name, int priority,
-                                  int process_set_id) {
+extern "C" int horovod_mxnet_reducescatter_async(NDArray* const* inputs,
+                                                 NDArray* const* outputs,
+                                                 const char* name, int priority,
+                                                 int process_set_id,
+                                                 int num_tensors) {
   MX_API_BEGIN();
 
 #if HAVE_CUDA && !HOROVOD_GPU_REDUCESCATTER
-  if (IsTensorOnCPU(input) && IsTensorOnCPU(output)) {
-    PushHorovodOperation(OperationType::REDUCESCATTER, &input, &output, name,
-                         priority, 1, process_set_id, -1, false);
+  if (IsTensorOnCPU(inputs[0]) && IsTensorOnCPU(outputs[0])) {
+    PushHorovodOperation(OperationType::REDUCESCATTER, inputs, outputs, name,
+                         priority, num_tensors, process_set_id, -1, ReduceOp::SUM);
   } else {
-    PushHorovodOperationCudaOnCPU(OperationType::REDUCESCATTER, &input, &output,
-                                  name, priority, 1, process_set_id, -1, false);
+    PushHorovodOperationCudaOnCPU(OperationType::REDUCESCATTER, inputs, outputs,
+                                  name, priority, num_tensors, process_set_id, -1, ReduceOp::SUM);
   }
 #else
-  PushHorovodOperation(OperationType::REDUCESCATTER, &input, &output, name,
-                       priority, 1, process_set_id, -1, false);
+  PushHorovodOperation(OperationType::REDUCESCATTER, inputs, outputs, name,
+                       priority, num_tensors, process_set_id, -1, ReduceOp::SUM);
 #endif
 
   MX_API_END();
