@@ -21,7 +21,7 @@ import numbers
 import os
 import time
 import warnings
-from distutils.version import LooseVersion
+from packaging import version
 
 from pyspark import keyword_only
 from pyspark.ml.param.shared import Param, Params, TypeConverters
@@ -135,17 +135,18 @@ class TorchEstimator(HorovodEstimator, TorchEstimatorParamsWritable,
                     for training. Not needed for lightning model.
         partitions_per_process: (Optional) Number of Parquet partitions to assign per worker
                                 process from `num_proc` (default: 10).
-        reader_pool_type:   (Optional) Type of worker pool used to parallelize reading data from
-                            the dataset. Should be one of ['thread', 'process']. Defaults to
-                            'process'.
+        reader_pool_type:   (Optional) Type of Petastorm worker pool used to parallelize reading data from
+                            the dataset. Should be one of ['thread', 'process', 'dummy']. Defaults to
+                            'thread'.
         run_id:     (Optional) unique ID for this run for organization in the Store. Will be
                     automatically assigned if not provided.
         sample_weight_col:  (Optional) column indicating the weight of each sample.
         random_seed: Optional random seed to use for PyTorch Lightning. Default: None.
-        shuffle_buffer_size: Optional size of in-memory shuffle buffer in rows (on training data).
+        shuffle_buffer_size: (Deprecated) Optional size of in-memory shuffle buffer in rows (on training data).
                              Allocating a larger buffer size increases randomness of shuffling at
                              the cost of more host memory. Defaults to estimating with an assumption
                              of 4GB of memory per host. Set shuffle_buffer_size=0 would turn off shuffle.
+        shuffle: (Optional) Whether to shuffle training samples or not. Defaults to True.
         store:      Store object that abstracts reading and writing of intermediate data and
                     run results.
         terminate_on_nan : (Optinoal) terminate the training process on seeing NaN output.
@@ -169,6 +170,12 @@ class TorchEstimator(HorovodEstimator, TorchEstimatorParamsWritable,
                             [TransformSpec](https://github.com/uber/petastorm/blob/master/petastorm/transform.py)
                             for more details. Note that this fucntion constructs another function
                             which should perform the transformation.
+        transformation_edit_fields: (Optional) A list of 4-tuples with the following fields:
+                            ``(name, numpy_dtype, shape, is_nullable)`` used for Petastorm 
+                            [TransformSpec](https://github.com/uber/petastorm/blob/master/petastorm/transform.py)
+                            to add more fields into the schema.
+        transformation_removed_fields: (Optional). A list of field names that will be removed from the original schema
+                            used for Petastorm [TransformSpec](https://github.com/uber/petastorm/blob/master/petastorm/transform.py).
         val_batch_size: Number of rows from the DataFrame per batch for validation, if not set,
                          will use batch_size.
         val_reader_num_workers: Similar to the train_reader_num_workers.
@@ -243,12 +250,15 @@ class TorchEstimator(HorovodEstimator, TorchEstimatorParamsWritable,
                  verbose=1,
                  random_seed=None,
                  shuffle_buffer_size=None,
+                 shuffle=True,
                  partitions_per_process=None,
                  run_id=None,
                  train_minibatch_fn=None,
                  train_steps_per_epoch=None,
                  validation_steps_per_epoch=None,
                  transformation_fn=None,
+                 transformation_edit_fields=None,
+                 transformation_removed_fields=None,
                  train_reader_num_workers=None,
                  trainer_args=None,
                  val_reader_num_workers=None,
@@ -288,7 +298,7 @@ class TorchEstimator(HorovodEstimator, TorchEstimatorParamsWritable,
         kwargs = self._input_kwargs
 
         # pl version check
-        if LooseVersion(pl.__version__) < LooseVersion(MIN_PL_VERSION):
+        if version.parse(pl.__version__) < version.parse(MIN_PL_VERSION):
             raise RuntimeError("Only support pytorch_lightning version > {}, found version {}".format(MIN_PL_VERSION, pl.__version__))
 
         if EstimatorParams.loss.name in kwargs and TorchEstimator.loss_constructors.name in kwargs:
