@@ -16,6 +16,7 @@
 # ==============================================================================
 # pylint: disable=g-short-docstring-punctuation
 
+from packaging import version
 import os
 import warnings
 
@@ -26,7 +27,8 @@ check_extension('horovod.tensorflow', 'HOROVOD_WITH_TENSORFLOW', __file__, 'mpi_
 from horovod.tensorflow import elastic
 from horovod.tensorflow.compression import Compression
 from horovod.tensorflow.functions import allgather_object, broadcast_object, broadcast_object_fn, broadcast_variables
-from horovod.tensorflow.mpi_ops import allgather, broadcast, broadcast_, _allreduce, _grouped_allreduce, alltoall, _reducescatter
+from horovod.tensorflow.mpi_ops import allgather, grouped_allgather, broadcast, broadcast_, _allreduce, \
+    _grouped_allreduce, alltoall, _reducescatter, _grouped_reducescatter
 from horovod.tensorflow.mpi_ops import init, shutdown
 from horovod.tensorflow.mpi_ops import is_initialized, start_timeline, stop_timeline
 from horovod.tensorflow.mpi_ops import size, local_size, cross_size, rank, local_rank, cross_rank, is_homogeneous
@@ -35,7 +37,7 @@ from horovod.tensorflow.mpi_ops import mpi_threads_supported, mpi_enabled, mpi_b
 from horovod.tensorflow.mpi_ops import gloo_enabled, gloo_built
 from horovod.tensorflow.mpi_ops import nccl_built, ddl_built, ccl_built, cuda_built, rocm_built
 from horovod.tensorflow.mpi_ops import ProcessSet, global_process_set, add_process_set, remove_process_set
-from horovod.tensorflow.mpi_ops import Average, Sum, Adasum
+from horovod.tensorflow.mpi_ops import Average, Sum, Adasum, Min, Max, Product
 from horovod.tensorflow.mpi_ops import handle_average_backwards_compatibility, check_num_rank_power_of_2
 from horovod.tensorflow.util import _executing_eagerly, _make_subgraph, _cache, vars_to_refs, refs_to_vars
 from horovod.tensorflow.mpi_ops import join
@@ -43,6 +45,7 @@ from horovod.tensorflow.sync_batch_norm import SyncBatchNormalization
 from horovod.tensorflow.gradient_aggregation import LocalGradientAggregationHelper
 
 import tensorflow as tf
+_IS_TF2 = version.parse(tf.__version__) >= version.parse('2.0.0')
 
 # @DEKHTIARJonathan: Do not remove, this fixes issues:
 # - https://github.com/tensorflow/tensorflow/issues/38516
@@ -55,7 +58,8 @@ if tf.__version__.startswith('2.2.'):
 def allreduce(tensor, average=None, device_dense='', device_sparse='',
               compression=Compression.none, op=None,
               prescale_factor=1.0, postscale_factor=1.0,
-              name=None, process_set=global_process_set):
+              name=None, process_set=global_process_set,
+              ignore_name_scope=False):
     """Perform an allreduce on a tf.Tensor or tf.IndexedSlices.
 
     This function performs a bandwidth-optimal ring allreduce on the input
@@ -79,12 +83,15 @@ def allreduce(tensor, average=None, device_dense='', device_sparse='',
                      sent and received by each worker node.  Defaults to not
                      using compression.
         op: The reduction operation to combine tensors across different ranks.
-            Defaults to Average if None is given.
+            Supported op values are Sum, Average, Min, Max, and Product. Defaults
+            to Average if None is given.
         prescale_factor: Multiplicative factor to scale tensor before allreduce.
         postscale_factor: Multiplicative factor to scale tensor after allreduce.
         process_set: Process set object to limit this operation to a subset of
             Horovod processes. Default is the global process set.
         name: A name of the allreduce operation
+        ignore_name_scope: If True, ignores any outer name scope applied by
+                           TensorFlow in the name used by the Horovod operation.
 
     Returns:
         A tensor of the same shape and type as `tensor`, summed across all
@@ -102,8 +109,8 @@ def allreduce(tensor, average=None, device_dense='', device_sparse='',
             horovod_size = tf.cast(size_op(process_set_id=process_set.process_set_id)
                                    if int(os.environ.get("HOROVOD_ELASTIC", 0)) else process_set.size(),
                                    dtype=tensor.values.dtype)
-            values = allgather(tensor.values, process_set=process_set)
-            indices = allgather(tensor.indices, process_set=process_set)
+            values = allgather(tensor.values, process_set=process_set, ignore_name_scope=ignore_name_scope)
+            indices = allgather(tensor.indices, process_set=process_set, ignore_name_scope=ignore_name_scope)
 
             # To make this operation into an average, divide allgathered values by
             # the Horovod size.
@@ -125,7 +132,8 @@ def allreduce(tensor, average=None, device_dense='', device_sparse='',
             summed_tensor_compressed = _allreduce(tensor_compressed, op=op,
                                                   prescale_factor=prescale_factor,
                                                   postscale_factor=postscale_factor,
-                                                  name=name, process_set=process_set)
+                                                  name=name, process_set=process_set,
+                                                  ignore_name_scope=ignore_name_scope)
             summed_tensor = compression.decompress(summed_tensor_compressed, ctx)
             if op == Adasum:
                 if process_set != global_process_set:
@@ -162,7 +170,8 @@ def allreduce(tensor, average=None, device_dense='', device_sparse='',
 
 
 def reducescatter(tensor, device_dense='', compression=Compression.none, op=Average,
-                  name=None, process_set=global_process_set):
+                  name=None, process_set=global_process_set,
+                  ignore_name_scope=False):
     """Perform a reducescatter on a tf.Tensor.
 
     This function performs a bandwidth-optimal reduce and scatter on the input
@@ -181,6 +190,8 @@ def reducescatter(tensor, device_dense='', compression=Compression.none, op=Aver
         process_set: Process set object to limit this operation to a subset of
             Horovod processes. Default is the global process set.
         name: A name of the reduce_scatter operation
+        ignore_name_scope: If True, ignores any outer name scope applied by
+                           TensorFlow in the name used by the Horovod operation.
 
     Returns:
         A tensor of the same rank and type as `tensor`, summed across all processes.
@@ -195,15 +206,52 @@ def reducescatter(tensor, device_dense='', compression=Compression.none, op=Aver
                                if int(os.environ.get("HOROVOD_ELASTIC", 0)) else process_set.size(),
                                dtype=tensor.dtype)
         tensor_compressed, ctx = compression.compress(tensor)
-        reduced_tensor_compressed = _reducescatter(tensor_compressed, op=true_op, name=name, process_set=process_set)
+        reduced_tensor_compressed = _reducescatter(tensor_compressed, op=true_op, name=name, process_set=process_set,
+                                                   ignore_name_scope=ignore_name_scope)
         reduced_tensor = compression.decompress(reduced_tensor_compressed, ctx)
         new_tensor = (reduced_tensor / horovod_size) if op == Average else reduced_tensor
     return new_tensor
 
+
 def grouped_allreduce(tensors, average=None, device_dense='', device_sparse='',
                       compression=Compression.none, op=None,
                       prescale_factor=1.0, postscale_factor=1.0,
-                      process_set=global_process_set):
+                      process_set=global_process_set,
+                      ignore_name_scope=False,
+                      name=None):
+    """Perform grouped allreduces on a sequence of tf.Tensor or tf.IndexedSlices.
+
+    Arguments:
+        tensors: Sequence of tf.Tensor, tf.Variable, or tf.IndexedSlices to reduce.
+                 The tensor type and shape must be the same on all Horovod processes
+                 for tensors sharing positions in `tensors`.
+        average:
+            .. warning:: .. deprecated:: 0.19.0
+
+                Use `op` instead. Will be removed in v1.0.
+
+        device_dense: Device to be used for dense tensors. Uses GPU by default
+                      if Horovod was built with HOROVOD_GPU_OPERATIONS.
+        device_sparse: Device to be used for sparse tensors. Uses GPU by default
+                       if Horovod was built with HOROVOD_GPU_OPERATIONS.
+        compression: Compression algorithm used to reduce the amount of data
+                     sent and received by each worker node.  Defaults to not
+                     using compression.
+        op: The reduction operation to combine tensors across different ranks.
+            Supported op values are Sum, Average, Min, Max, and Product. Defaults
+            to Average if None is given.
+        prescale_factor: Multiplicative factor to scale tensors before allreduce.
+        postscale_factor: Multiplicative factor to scale tensors after allreduce.
+        process_set: Process set object to limit this operation to a subset of
+            Horovod processes. Default is the global process set.
+        name: A name of the reduce_scatter operation
+        ignore_name_scope: If True, ignores any outer name scope applied by
+                           TensorFlow in the name used by the Horovod operation.
+
+    Returns:
+        A list of tensors of the same shape and type as those in `tensors`,
+        reduced across all processes.
+    """
     if not tensors:
         return tensors
 
@@ -227,8 +275,8 @@ def grouped_allreduce(tensors, average=None, device_dense='', device_sparse='',
                 horovod_size = tf.cast(size_op(process_set_id=process_set.process_set_id)
                                        if int(os.environ.get("HOROVOD_ELASTIC", 0)) else process_set.size(),
                                        dtype=tensor.values.dtype)
-                values = allgather(tensor.values, process_set=process_set)
-                indices = allgather(tensor.indices, process_set=process_set)
+                values = allgather(tensor.values, process_set=process_set, ignore_name_scope=ignore_name_scope)
+                indices = allgather(tensor.indices, process_set=process_set, ignore_name_scope=ignore_name_scope)
 
                 # To make this operation into an average, divide allgathered values by
                 # the Horovod size.
@@ -241,7 +289,9 @@ def grouped_allreduce(tensors, average=None, device_dense='', device_sparse='',
             summed_tensors_compressed = _grouped_allreduce(tensors_compressed, op=op,
                                                            prescale_factor=prescale_factor,
                                                            postscale_factor=postscale_factor,
-                                                           process_set=process_set)
+                                                           process_set=process_set,
+                                                           name=name,
+                                                           ignore_name_scope=ignore_name_scope)
             summed_tensors = [compression.decompress(t, ctx) for t, ctx in zip(summed_tensors_compressed, ctxs)]
             if op == Adasum:
                 if process_set != global_process_set:
@@ -283,6 +333,7 @@ def grouped_allreduce(tensors, average=None, device_dense='', device_sparse='',
                     new_tensors = summed_tensors
         return new_tensors
 
+
 def _allreduce_cond(tensor, *args, process_set=global_process_set, **kwargs):
     def allreduce_fn():
         return allreduce(tensor, *args, process_set=process_set, **kwargs)
@@ -297,6 +348,7 @@ def _allreduce_cond(tensor, *args, process_set=global_process_set, **kwargs):
         tf.convert_to_tensor(process_set.included() and process_set.size() > 1)),
                    allreduce_fn, id_fn)
 
+
 def _grouped_allreduce_cond(tensors, *args, process_set=global_process_set, **kwargs):
     def allreduce_fn():
         return grouped_allreduce(tensors, *args, process_set=process_set, **kwargs)
@@ -312,6 +364,48 @@ def _grouped_allreduce_cond(tensors, *args, process_set=global_process_set, **kw
                    allreduce_fn, id_fn)
 
 
+def grouped_reducescatter(tensors, device_dense='', compression=Compression.none, op=Average,
+                          process_set=global_process_set):
+    """Perform grouped reducescatters on a sequence of tf.Tensor.
+
+    Arguments:
+        tensors: Sequence of tf.Tensor or tf.Variable to reduce.
+                 The shape must be the same on all Horovod processes
+                 for inputs sharing positions in `tensors`.
+        device_dense: Device to be used for dense tensors. Uses GPU by default
+                      if Horovod was built with HOROVOD_GPU_OPERATIONS.
+        device_sparse: Device to be used for sparse tensors. Uses GPU by default
+                       if Horovod was built with HOROVOD_GPU_OPERATIONS.
+        compression: Compression algorithm used to reduce the amount of data
+                     sent and received by each worker node.  Defaults to not
+                     using compression.
+        op: The reduction operation to combine tensors across different ranks.
+            Defaults to Average if None is given.
+        process_set: Process set object to limit this operation to a subset of
+            Horovod processes. Default is the global process set.
+
+    Returns:
+        A list of tensors of the same rank and type as those in `tensors`,
+        reduced across all processes. For each returned tensor the shape is
+        identical to the corresponding input shape, except for the first
+        dimension, which will be divided across the different Horovod processes.
+    """
+    if not tensors:
+        return tensors
+    # Averaging happens in framework code, so translate that to Sum for the actual call
+    true_op = Sum if op == Average else op
+    dtype = tensors[0].dtype  # HorovodGroupedReducescatterOp requires all input tensors to have the same dtype
+    with tf.device(device_dense):
+        horovod_size = tf.cast(size_op(process_set_id=process_set.process_set_id)
+                               if int(os.environ.get("HOROVOD_ELASTIC", 0)) else process_set.size(),
+                               dtype=dtype)
+        tensors_compressed, ctxs = zip(*[compression.compress(tensor) for tensor in tensors])
+        reduced_tensors_compressed = _grouped_reducescatter(tensors_compressed, op=true_op, process_set=process_set)
+        reduced_tensors = [compression.decompress(t, ctx) for t, ctx in zip(reduced_tensors_compressed, ctxs)]
+        new_tensors = [(rt / horovod_size) for rt in reduced_tensors] if op == Average else reduced_tensors
+    return new_tensors
+
+
 try:
     _global_variables = tf.compat.v1.global_variables
 except AttributeError:
@@ -406,7 +500,7 @@ def _make_cached_allreduce_grads_fn(name, device_dense, device_sparse,
         prescale_factor = 1.0
         postscale_factor = 1.0
 
-    def allreduce_grads(grads, vars=None):
+    def allreduce_grads(grads, vars=None, use_generic_names=False):
         with tf.name_scope(name + "_Allreduce"):
             if sparse_as_dense:
                 grads = [tf.convert_to_tensor(grad)
@@ -436,7 +530,7 @@ def _make_cached_allreduce_grads_fn(name, device_dense, device_sparse,
                     grads_split = split_list(grads_clean, groups)
 
                 reduce_ops = [None] * len(vars)
-                for group in grads_split:
+                for i, group in enumerate(grads_split):
                     index_group, grad_group = [list(t) for t in zip(*group)]
                     reduce_ops_group = _grouped_allreduce_cond(grad_group,
                                                                device_dense=device_dense,
@@ -445,7 +539,9 @@ def _make_cached_allreduce_grads_fn(name, device_dense, device_sparse,
                                                                op=op,
                                                                prescale_factor=prescale_factor,
                                                                postscale_factor=postscale_factor,
-                                                               process_set=process_set)
+                                                               process_set=process_set,
+                                                               name=f"grad_{i}" if use_generic_names else None,
+                                                               ignore_name_scope=use_generic_names)
                     for i in range(len(index_group)):
                         reduce_ops[index_group[i]] = reduce_ops_group[i]
                 return reduce_ops
@@ -457,9 +553,11 @@ def _make_cached_allreduce_grads_fn(name, device_dense, device_sparse,
                                     op=op,
                                     prescale_factor=prescale_factor,
                                     postscale_factor=postscale_factor,
-                                    process_set=process_set)
+                                    process_set=process_set,
+                                    name=f"grad_{i}" if use_generic_names else None,
+                                    ignore_name_scope=use_generic_names)
                     if grad is not None else grad
-                    for grad in grads]
+                    for i, grad in enumerate(grads)]
 
     if _executing_eagerly():
         return _make_subgraph(allreduce_grads)
@@ -498,7 +596,7 @@ if _LegacyOptimizer is not None:
                     device_sparse='', compression=Compression.none,
                     sparse_as_dense=False, op=Average, gradient_predivide_factor=1.0,
                     backward_passes_per_step=1, average_aggregated_gradients=False,
-                    groups=None, process_set=global_process_set):
+                    groups=None, process_set=global_process_set, scale_local_gradients=True):
             if name is None:
                 name = "Distributed{}".format(type(optimizer).__name__)
             super(_DistributedOptimizer, self).__init__(name=name, use_locking=use_locking)
@@ -508,6 +606,9 @@ if _LegacyOptimizer is not None:
                 name, device_dense, device_sparse, compression, sparse_as_dense, op,
                 gradient_predivide_factor, groups, process_set=process_set)
 
+            self._local_vars = set()
+            self.process_set = process_set
+            self.scale_local_gradients = scale_local_gradients
             self._agg_helper = None
             if backward_passes_per_step > 1:
                 if _executing_eagerly():
@@ -523,8 +624,21 @@ if _LegacyOptimizer is not None:
                     average_aggregated_gradients=average_aggregated_gradients,
                     rank=rank(),
                     optimizer_type=LocalGradientAggregationHelper._OPTIMIZER_TYPE_LEGACY,
+                    process_set=process_set,
+                    scale_local_gradients=scale_local_gradients
                 )
 
+        def register_local_var(self, var):
+            """Registers a source/variable as worker local. Horovod will not perform any global
+            operations on gradients corresponding to these sources and will instead return the local
+            gradient."""    
+            if self._agg_helper:
+                self._agg_helper.register_local_var(var)
+            elif _IS_TF2:
+                self._local_vars.add(var.ref())
+            else:
+                self._local_vars.add(var)
+
         def compute_gradients(self, *args, **kwargs):
             """Compute gradients of all trainable variables.
 
@@ -538,7 +652,48 @@ if _LegacyOptimizer is not None:
             if self._agg_helper:
                 avg_grads = self._agg_helper.compute_gradients(grads, vars)
             else:
-                avg_grads = self._allreduce_grads(grads, vars)
+                def _filtered_reduce_grads(grads, vars):
+                    rv = []
+                    rg = []
+                    if _IS_TF2:
+                        v2g = {var.ref(): grad for var, grad in zip(vars, grads)}
+                        for var, grad in zip(vars, grads):
+                            if var.ref() not in self._local_vars:
+                                rv.append(var)
+                                rg.append(grad)
+                    else:
+                        v2g = {var: grad for var, grad in zip(vars, grads)}
+                        for var, grad in zip(vars, grads):
+                            if var not in self._local_vars:
+                                rv.append(var)
+                                rg.append(grad)
+
+                    rg = self._allreduce_grads(rg, rv)
+                    horovod_size = size_op(process_set_id=self.process_set.process_set_id) if int(os.environ.get("HOROVOD_ELASTIC", 0)) else self.process_set.size()
+                    if _IS_TF2:
+                        for rv,rg in zip(rv, rg):
+                            v2g[rv.ref()] = rg
+
+                        if self.scale_local_gradients and len(self._local_vars):
+                            # Scale local gradients by a size factor. See pull/3695 and discussions/3705 for context.
+                            for v_ref in v2g:
+                                if v_ref in self._local_vars and v2g[v_ref]:
+                                    v2g[v.ref()] /= horovod_size
+
+                        return [v2g[rv.ref()] for rv in vars]
+                    else:
+                        for rv, rg in zip(rv, rg):
+                            v2g[rv] = rg
+
+                        if self.scale_local_gradients and len(self._local_vars):
+                            # Scale local gradients by a size factor. See pull/3695 and discussions/3705 for context.
+                            for v in v2g:
+                                if v in self._local_vars and v2g[v]:
+                                    v2g[v] /= horovod_size
+
+                        return [v2g[rv] for rv in vars]
+
+                avg_grads = _filtered_reduce_grads(grads, vars)
             return list(zip(avg_grads, vars))
 
         def apply_gradients(self, grads_and_vars, global_step=None, name=None):
@@ -669,7 +824,8 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
                          sparse_as_dense=False, backward_passes_per_step=1,
                          op=Average, gradient_predivide_factor=1.0,
                          average_aggregated_gradients=False,
-                         num_groups=0, groups=None, process_set=global_process_set):
+                         num_groups=0, groups=None,
+                         process_set=global_process_set, scale_local_gradients=True):
     """Construct a new DistributedOptimizer, which uses another optimizer
     under the hood for computing single-process gradient values and
     applying gradient updates after the gradient values have been combined
@@ -729,6 +885,7 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
         Defaults as None, which is no explicit groups.
       process_set: Gradients will only be reduced over Horovod processes belonging
         to this process set. Defaults to the global process set.
+      scale_local_gradients: Whether to scale the gradients of local variables. Default is set to True.
     """
     if gradient_predivide_factor != 1.0:
         if rocm_built():
@@ -771,6 +928,7 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
             average_aggregated_gradients=average_aggregated_gradients,
             groups=groups,
             process_set=process_set,
+            scale_local_gradients=scale_local_gradients
         )
     elif isinstance(optimizer, tf.keras.optimizers.Optimizer):
         if op == Adasum:
@@ -788,6 +946,7 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
             backward_passes_per_step=backward_passes_per_step,
             average_aggregated_gradients=average_aggregated_gradients,
             process_set=process_set,
+            scale_local_gradients=scale_local_gradients
         )
     else:
         raise ValueError('Provided optimizer doesn\'t inherit from either legacy '
@@ -798,7 +957,7 @@ if hasattr(tf, 'GradientTape'):
     class _DistributedGradientTape(tf.GradientTape):
         def __init__(self, tape, device_dense, device_sparse, compression, sparse_as_dense, op,
                      gradient_predivide_factor, groups, persistent=False,
-                     watch_accessed_variables=True, process_set=global_process_set):
+                     watch_accessed_variables=True, process_set=global_process_set, scale_local_gradients=True):
             if hasattr(tape, '_watch_accessed_variables'):
                 super(self.__class__, self).__init__(persistent, watch_accessed_variables)
             else:
@@ -809,15 +968,70 @@ if hasattr(tf, 'GradientTape'):
                 'DistributedGradientTape', device_dense, device_sparse, compression,
                 sparse_as_dense, op, gradient_predivide_factor, groups, process_set)
 
-        def gradient(self, target, sources, output_gradients=None):
-            gradients = super(self.__class__, self).gradient(target, sources, output_gradients)
-            return self._allreduce_grads(gradients, sources)
+            self.process_set = process_set
+            self.scale_local_gradients = scale_local_gradients
+            self._local_sources = set()
 
 
+        def register_local_source(self, source):
+            """Registers a source/variable as worker local. Horovod will not perform any global
+            operations on gradients corresponding to these sources and will instead return the local
+            gradient."""
+            if _IS_TF2:
+                self._local_sources.add(source.ref())
+            else:
+                self._local_sources.add(source)
+
+        def gradient(self, target, sources, output_gradients=None, use_generic_names=False):
+            gradients = super(self.__class__, self).gradient(target, sources, output_gradients)
+
+            # Collect source/grad pairs requiring reduction (i.e. not from a registered local source)
+            rs = []
+            rg = []
+            if _IS_TF2:
+                s2g = {s.ref() : g for s,g in zip(sources, gradients)}
+                for s,g in zip(sources, gradients):
+                    if s.ref() not in self._local_sources:
+                        rs.append(s)
+                        rg.append(g)
+            else:
+                s2g = {s : g for s,g in zip(sources, gradients)}
+                for s,g in zip(sources, gradients):
+                    if s not in self._local_sources:
+                        rs.append(s)
+                        rg.append(g)
+
+            # Reduce grads
+            rg = self._allreduce_grads(rg, rs, use_generic_names)
+            horovod_size = size_op(process_set_id=self.process_set.process_set_id) if int(os.environ.get("HOROVOD_ELASTIC", 0)) else self.process_set.size()
+            # Replace dict entries with reduced grads
+            if _IS_TF2:
+                for rs, rg in zip(rs, rg):
+                    s2g[rs.ref()] = rg
+
+                if self.scale_local_gradients and len(self._local_sources):
+                    # Scale local gradients by a size factor. See pull/3695 and discussions/3705 for context.
+                    for s_ref in s2g:
+                        if s_ref in self._local_sources and s2g[s_ref] is not None:
+                            s2g[s_ref] /= horovod_size
+
+                return [s2g[s.ref()] for s in sources]
+            else:
+                for rs, rg in zip(rs, rg):
+                    s2g[rs] = rg
+
+                if self.scale_local_gradients and len(self._local_sources):
+                    # Scale local gradients by a size factor. See pull/3695 and discussions/3705 for context.
+                    for s in s2g:
+                        if s in self._local_sources and s2g[s] is not None:
+                            s2g[s] /= horovod_size
+
+                return [s2g[s] for s in sources]
+
     def DistributedGradientTape(gradtape, device_dense='', device_sparse='',
                                 compression=Compression.none, sparse_as_dense=False,
                                 op=Average, gradient_predivide_factor=1.0,
-                                num_groups=0, groups=None, process_set=global_process_set):
+                                num_groups=0, groups=None, process_set=global_process_set, scale_local_gradients=True):
         """A tape that wraps another tf.GradientTape, using an allreduce to
         combine gradient values before applying gradients to model weights.
 
@@ -860,6 +1074,7 @@ if hasattr(tf, 'GradientTape'):
             Defaults as None, which is no explicit groups.
           process_set: Gradients will only be reduced over Horovod processes belonging
             to this process set. Defaults to the global process set.
+          scale_local_gradients: Whether to scale the gradients of local variables. Default is set to True.
         """
         if gradient_predivide_factor != 1.0:
             if rocm_built():
@@ -880,12 +1095,52 @@ if hasattr(tf, 'GradientTape'):
 
         cls = type(gradtape.__class__.__name__, (gradtape.__class__,),
                    dict(_DistributedGradientTape.__dict__))
+
         if hasattr(gradtape, '_watch_accessed_variables'):
             return cls(gradtape._tape, device_dense, device_sparse, compression,
                        sparse_as_dense, op, gradient_predivide_factor, groups,
                        gradtape._persistent, gradtape._watch_accessed_variables,
-                       process_set=process_set)
+                       process_set=process_set, scale_local_gradients=scale_local_gradients)
         else:
             return cls(gradtape._tape, device_dense, device_sparse, compression,
                        sparse_as_dense, op, gradient_predivide_factor, groups,
-                       gradtape._persistent, process_set=process_set)
+                       gradtape._persistent, process_set=process_set, scale_local_gradients=scale_local_gradients)
+
+
+    def PartialDistributedGradientTape(gradtape, device_dense='', device_sparse='',
+                                       compression=Compression.none, sparse_as_dense=False,
+                                       op=Average, gradient_predivide_factor=1.0,
+                                       num_groups=0, groups=None, process_set=global_process_set,
+                                       local_layers=None, scale_local_gradients=True):
+        """A tape that wraps another tf.GradientTape, using an allreduce to
+        combine gradient values before applying gradients to model weights similar to
+        DistributedGradientTape execpt it skips allreducing gradients of the local layers
+        passed in local_layers parameter.
+
+        Args:
+          gradtape:
+            GradientTape to use for computing gradients and applying updates.
+          local_layers:
+            A collection of type tf.keras.layers.Layer local layers that their gradients need not
+            to be synced accross ranks and is kept and applied locally.
+            If not provided, the functionality of PartialDistributedGradientTape is
+            identical to DistributedGradientTape.
+
+        The rest of the arguments are similar to those of DistributedGradientTape.
+        """
+        if local_layers is None:
+            local_layers = []
+        elif isinstance(local_layers, tf.keras.layers.Layer):
+            local_layers = [local_layers]
+        elif not all(isinstance(layer, tf.keras.layers.Layer) for layer in local_layers):
+            raise ValueError("All local layers must be of tf.keras.layers.Layer type.")
+
+        local_vars = [var for layer in local_layers for var in layer.trainable_weights]
+
+        _tape = DistributedGradientTape(gradtape, device_dense, device_sparse,
+                                        compression, sparse_as_dense,
+                                        op, gradient_predivide_factor,
+                                        num_groups, groups, process_set, scale_local_gradients)
+        for var in local_vars:
+            _tape.register_local_source(var)
+        return _tape
