@@ -22,6 +22,7 @@ from tensorflow import keras
 
 from horovod.common.util  import is_version_greater_equal_than
 
+
 if is_version_greater_equal_than(tf.__version__, "2.6.0"):
     from keras import backend as K
 else:
@@ -62,7 +63,8 @@ def DistributedOptimizer(optimizer, name=None,
                          average_aggregated_gradients=False,
                          num_groups=0,
                          groups=None,
-                         process_set=global_process_set):
+                         process_set=global_process_set,
+                         scale_local_gradients=True):
     """
     An optimizer that wraps another keras.optimizers.Optimizer, using an allreduce to
     average gradient values before applying gradients to model weights.
@@ -107,8 +109,10 @@ def DistributedOptimizer(optimizer, name=None,
                 inner list will be assigned to the same group, while parameter that does
                 not appear in any list will form a group itself.
                 Defaults as None, which is no explicit groups.
-      process_set: Gradients will only be reduced over Horovod processes belonging
+        process_set: Gradients will only be reduced over Horovod processes belonging
                    to this process set. Defaults to the global process set.
+        scale_local_gradients: Whether to scale the gradients of local variables. Default is set to True.
+
     """
     if gradient_predivide_factor != 1.0 and rocm_built():
             raise ValueError('gradient_predivide_factor not supported yet with ROCm')
@@ -141,6 +145,7 @@ def DistributedOptimizer(optimizer, name=None,
         average_aggregated_gradients=average_aggregated_gradients,
         groups=groups,
         process_set=process_set,
+        scale_local_gradients=scale_local_gradients
     )
 
 
