@@ -14,12 +14,8 @@
 // =============================================================================
 
 #if HAVE_GPU
-#if TORCH_VERSION >= 1005000000
 #include <c10/cuda/CUDAStream.h>
 #include <c10/cuda/CUDAException.h>
-#else
-#include <THC/THC.h>
-#endif
 #include <cassert>
 #include <mutex>
 #include <queue>
@@ -31,12 +27,6 @@
 #include "ready_event.h"
 #include "cuda_util.h"
 
-#if TORCH_VERSION < 1005000000
-#if HAVE_GPU
-extern THCState* state;
-#endif
-#endif
-
 namespace horovod {
 namespace torch {
 
@@ -59,22 +49,12 @@ TorchReadyEvent::TorchReadyEvent(int device) : device_(device) {
       cuda_event_ = queue.front();
       queue.pop();
     } else {
-      #if TORCH_VERSION >= 1005000000
       C10_CUDA_CHECK(cudaEventCreateWithFlags(
           &cuda_event_, cudaEventBlockingSync | cudaEventDisableTiming));
-      #else
-      THCudaCheck(cudaEventCreateWithFlags(
-          &cuda_event_, cudaEventBlockingSync | cudaEventDisableTiming));
-      #endif
     }
   }
-  #if TORCH_VERSION >= 1005000000
   auto stream = c10::cuda::getCurrentCUDAStream(device_);
   C10_CUDA_CHECK(cudaEventRecord(cuda_event_, stream));
-  #else
-  auto stream = THCState_getCurrentStreamOnDevice(state, device_);
-  THCudaCheck(cudaEventRecord(cuda_event_, stream));
-  #endif
 }
 
 TorchReadyEvent::~TorchReadyEvent() {
@@ -86,17 +66,13 @@ TorchReadyEvent::~TorchReadyEvent() {
 }
 
 bool TorchReadyEvent::Ready() const {
-  auto status = cudaEventQuery(cuda_event_);
-  if (status == cudaErrorNotReady) {
-    return false;
-  }
-  #if TORCH_VERSION >= 1005000000
-  C10_CUDA_CHECK(status);
-  #else
-  THCudaCheck(status);
-  #endif
+  C10_CUDA_CHECK(cudaEventSynchronize(cuda_event_));
   return true;
 }
+
+gpuEvent_t TorchReadyEvent::event() const {
+  return cuda_event_;
+}
 #endif
 
 // On GPU this event will signal that GPU computations are done and data is
