@@ -35,10 +35,11 @@ public:
   TorchReadyEvent(int device);
   ~TorchReadyEvent();
   virtual bool Ready() const override;
+  gpuEvent_t event() const override;
 
 private:
   int device_ = CPU_DEVICE_ID;
-  cudaEvent_t cuda_event_ = nullptr;
+  gpuEvent_t cuda_event_ = nullptr;
 };
 #endif
 
